@@ -1,7 +1,8 @@
 import os
 import jinja2
 import yaml
-from collections import MutableMapping, defaultdict, Iterable
+import six
+from collections import MutableMapping, defaultdict, Iterable, Mapping
 from glob import glob
 import yaml.constructor
 try:
@@ -9,6 +10,10 @@ try:
 except ImportError:
     OrderedDict = None
 
+if six.PY2:
+    getcwd = lambda filename: os.getcwdu() if isinstance(filename, unicode) else os.getcwd()
+else:
+    getcwd = lambda filename: os.getcwd()
 
 if OrderedDict:
     class OrderedDictYAMLLoader(yaml.Loader):
@@ -19,8 +24,8 @@ if OrderedDict:
         def __init__(self, *args, **kwargs):
             yaml.Loader.__init__(self, *args, **kwargs)
 
-            self.add_constructor(u'tag:yaml.org,2002:map', type(self).construct_yaml_map)
-            self.add_constructor(u'tag:yaml.org,2002:omap', type(self).construct_yaml_map)
+            self.add_constructor('tag:yaml.org,2002:map', type(self).construct_yaml_map)
+            self.add_constructor('tag:yaml.org,2002:omap', type(self).construct_yaml_map)
 
         def construct_yaml_map(self, node):
             data = OrderedDict()
@@ -33,14 +38,15 @@ if OrderedDict:
                 self.flatten_mapping(node)
             else:
                 raise yaml.constructor.ConstructorError(None, None,
-                                                        'expected a mapping node, but found %s' % node.id, node.start_mark)
+                                                        'expected a mapping node, but found %s' % node.id,
+                                                        node.start_mark)
 
             mapping = OrderedDict()
             for key_node, value_node in node.value:
                 key = self.construct_object(key_node, deep=deep)
                 try:
                     hash(key)
-                except TypeError, exc:
+                except TypeError as exc:
                     raise yaml.constructor.ConstructorError('while constructing a mapping',
                                                             node.start_mark,
                                                             'found unacceptable key (%s)' % exc, key_node.start_mark)
@@ -58,8 +64,9 @@ class FileNotFound(MetaYamlException):
 
 
 def _to_str(value):
-    if isinstance(value, unicode):
-        return value
+    if six.PY2:
+        if isinstance(value, unicode):
+            return value
     return str(value)
 
 
@@ -71,6 +78,19 @@ class MetaYaml(object):
     DEL_ALL_MARKER = "${__del_all__}"
     EXTEND_MARKER = "${__extend__}"
 
+    def cp(self, source, *args, **kwargs):
+        if isinstance(source, Mapping):
+            result = source.copy()
+            for arg in args:
+                source.update(arg)
+            result.update(kwargs)
+            return result
+        elif isinstance(source, (list, tuple)):
+            result = list(source)
+            result.extend(args)
+            return result
+        raise MetaYamlException("cp method support only dict and list as source")
+
     def __init__(self, yaml_file, defaults=None, extend_key_word="extend",
                  ignore_errors=False, ignore_not_existed_files=False,
                  disable_order_dict=False):
@@ -88,17 +108,23 @@ class MetaYaml(object):
 
         self._extend_key_word = extend_key_word
         self.disable_order_dict = disable_order_dict or (not OrderedDict)
+        self.parsed_ids = set()
         if disable_order_dict:
             self.data = defaults or {}
         else:
             self.data = OrderedDict(defaults) if defaults else {}
+        self.data["cp"] = self.cp
 
         self.cache_template = defaultdict(lambda: {})
         self.ignore_errors = ignore_errors
         self.ignore_not_existed_files = ignore_not_existed_files
         self.processed_files = set()
+        self.eval_value = self._simple_eval_value
+        self.substitute(self.data, self.data, ("defaults:", ), False)
+        self.defaults_ids = self.parsed_ids.copy()
+        self.eval_value = self._eval_value
 
-        if isinstance(yaml_file, basestring):
+        if isinstance(yaml_file, six.string_types):
             yaml_file = [yaml_file]
         else:
             assert isinstance(yaml_file, Iterable), "yaml_file should be string or list of strings"
@@ -107,15 +133,18 @@ class MetaYaml(object):
         for filename in files:
             self.load(filename, self.data)
 
-        self.substitute(self.data, self.data, [os.path.basename(files[0])], False)
+        self.parsed_ids = self.defaults_ids
+        self.substitute(self.data, self.data, (os.path.basename(files[0]), ), False)
         self.data.pop(self._extend_key_word, None)
+        if self.data["cp"] == self.cp:
+            del self.data["cp"]
 
     def extend_filename(self, file_list, path=None):
         files = []
         for filename in file_list:
             if not os.path.isabs(filename):
                 if not path:
-                    path = os.getcwdu() if isinstance(filename, unicode) else os.getcwd()
+                    path = getcwd(filename)
 
                 filename = os.path.join(path, filename)
 
@@ -132,7 +161,7 @@ class MetaYaml(object):
             return data  # file was already processed
 
         self.processed_files.add(path)
-        basename = [os.path.basename(path)]
+        basename = (os.path.basename(path), )
         file_dir = os.path.dirname(path)
 
         with open(path, "rb") as f:
@@ -146,7 +175,7 @@ class MetaYaml(object):
 
         if extends:
             self.substitute(data, data, basename, eager=True)
-            if isinstance(extends, basestring):
+            if isinstance(extends, six.string_types):
                 extends = [extends]
 
             if not isinstance(extends, Iterable):
@@ -154,7 +183,7 @@ class MetaYaml(object):
                                         self._path_to_str([basename, self._extend_key_word]))
 
             for file_name in extends:
-                if not isinstance(file_name, basestring):
+                if not isinstance(file_name, six.string_types):
                     raise MetaYamlException("The value of %s should be list of string or string" %
                                             self._path_to_str([basename, self._extend_key_word]))
 
@@ -170,11 +199,11 @@ class MetaYaml(object):
         return data
 
     def substitute(self, value, data, path, eager):
-        path = path or []
+        path = path or tuple()
         if isinstance(value, MutableMapping):
             to_remove = []
-            for key, val in value.iteritems():
-                new_path = path + [_to_str(key)]
+            for key, val in six.iteritems(value):
+                new_path = path + (_to_str(key), )
                 new_key = self.eval_value(key, new_path, data, eager)
                 if new_key != key:
                     to_remove.append(key)
@@ -183,9 +212,9 @@ class MetaYaml(object):
             for k in to_remove:
                 del value[k]
         elif isinstance(value, list):
-            for key in xrange(len(value)):
-                value[key] = self.substitute(value[key], data, path + [_to_str(key)], eager)
-        elif isinstance(value, basestring):
+            for key in range(len(value)):
+                value[key] = self.substitute(value[key], data, path + (_to_str(key), ), eager)
+        elif isinstance(value, six.string_types):
             value = self.eval_value(value, path, data, eager)
 
         return value
@@ -193,7 +222,7 @@ class MetaYaml(object):
     @staticmethod
     def _path_to_str(path):
         def to_str(p):
-            if isinstance(p, basestring):
+            if isinstance(p, six.string_types):
                 # noinspection PyTypeChecker
                 return "." + p
             else:
@@ -205,11 +234,18 @@ class MetaYaml(object):
             s = s[1:]
         return s
 
+    def _simple_eval_value(self, val, path, data, eager):
+        self.parsed_ids.add(id(val))
+        return val
+
     # noinspection PyUnresolvedReferences
-    def eval_value(self, val, path, data, eager):
-        if not isinstance(val, basestring):
+    def _eval_value(self, val, path, data, eager):
+        if not isinstance(val, six.string_types):
+            return val
+        if id(val) in self.parsed_ids:
             return val
 
+        self.parsed_ids.add(id(val))
         brackets = self.eager_brackets if eager else self.lazy_brackets
         if brackets[0] not in val:
             return val
@@ -223,18 +259,19 @@ class MetaYaml(object):
         if t is None:
             try:
                 t = jinja2.Template(val, variable_start_string=brackets[0], variable_end_string=brackets[1])
-            except Exception, e:
+            except Exception as e:
                 if not self.ignore_errors:
                     raise MetaYamlException("Template compiling error of key: %s, value: %s, error: %s" % (
                         self._path_to_str(path), val, e))
             cache[val] = t
         try:
-            r = list(t.root_render_func(t.new_context(data)))
+            data_str_key = {_to_str(k): v for k, v in six.iteritems(data)}
+            r = list(t.root_render_func(t.new_context(data_str_key)))
             if len(r) == 1:
                 result = r[0]
             else:
-                r = [unicode(rr) for rr in r]
-                result = u"".join(r)
+                r = [_to_str(rr) for rr in r]
+                result = six.u("").join(r)
         except Exception as e:
             result = val
             if not self.ignore_errors:
@@ -242,7 +279,7 @@ class MetaYaml(object):
         finally:
             jinja2.runtime.to_string = original_to_string
 
-        if isinstance(result, basestring):
+        if isinstance(result, six.string_types):
             for t in [int, float]:
                 try:
                     result = t(result)
@@ -254,11 +291,11 @@ class MetaYaml(object):
 
     def _merge(self, a, b, data, path=None):
         if path is None:
-            path = []
-        for key, b_value in b.iteritems():
+            path = tuple()
+        for key, b_value in six.iteritems(b):
             if key == self.DEL_MARKER:
-                new_value = self.substitute(b_value, data, path + [str(key)], True)
-                if isinstance(new_value, Iterable) and not isinstance(new_value, basestring):
+                new_value = self.substitute(b_value, data, path + (str(key), ), True)
+                if isinstance(new_value, Iterable) and not isinstance(new_value, six.string_types):
                     for v in new_value:
                         a.pop(v, None)
                 else:
@@ -266,12 +303,12 @@ class MetaYaml(object):
             elif key == self.DEL_ALL_MARKER:
                 a.clear()
 
-        for key, b_value in b.iteritems():
+        for key, b_value in six.iteritems(b):
             if key in (self.DEL_MARKER, self.DEL_ALL_MARKER):
                 continue
             a_value = a.get(key)
             if isinstance(a_value, MutableMapping) and isinstance(b_value, MutableMapping):
-                self._merge(a[key], b_value, path + [str(key)])
+                self._merge(a[key], b_value, path + (str(key), ))
             elif (isinstance(a_value, list) and isinstance(b_value, MutableMapping) and
                   len(b_value) == 1 and self.EXTEND_MARKER in b_value):
                 extend = b_value[self.EXTEND_MARKER]
