@@ -53,6 +53,32 @@ class TestMetaYaml(TestCase):
         self.assertEqual(d["main"]["remove_all_from_here"], {8: 8})
         self.assertEqual(d["main"]["test1"], ["v3", "v4", "v5"])
 
+    def test_disable_order_dict(self):
+        d = read(self._file_name("test.yaml"), {"CWD": os.getcwd(), "join": os.path.join},
+                 disable_order_dict=True)
+        self.assertEqual(type(d), dict)
+
+    def test_cp(self):
+        d = read(self._file_name("cp.yaml"))
+        schedule = d["schedule"]
+        self.assertEqual(schedule["nighttask"]["min"], 5)
+        self.assertEqual(schedule["nighttask"]["hour"], 0)
+
+        self.assertEqual(schedule["daytask"]["min"], 7)
+        self.assertEqual(schedule["daytask"]["hour"], 13)
+
+        self.assertEqual(schedule["monthtask"]["min"], 0)
+        self.assertEqual(schedule["monthtask"]["day"], 2)
+
+        self.assertEqual(d["deploy"]["elb"], ["1.1.1.1", "2.2.2.2", "3.3.3.3", "4.4.4.4", "5.5.5.5"])
+
+    def test_not_parsible_defaults(self):
+        not_parsible = "\[\e]0;\u@\h: \w\a\]${debian_chroot:+($(debian_chroot)}\u@\h:\w\$"
+        d = read(self._file_name("test.yaml"),
+                 {"env": {"PS1": not_parsible},
+                  "join": os.path.join})
+        self.assertEqual(d["env"]["PS1"], not_parsible)
+
 
 if __name__ == '__main__':
     main()
